##< constants
MINREAD <- 10
EPSILONREAD <- MINREAD/2
TANDESCEND <- tan(89/180*pi)
PVTHRESHOLD <- 1e-2
LOCPVTHRESHOLD <- 1e-2
LOWCOVTHRESHOLD <- 0.05
MARGIN <- 10
MINREDUCEDSIZE <- 50
MAXREDUCEDSIZE <- 2000
LENTHRESHOLD <- 50
MINCANDSIZE <- ceiling(LENTHRESHOLD/2)
FIGUREPATH <- "./"
MINDOMINANTBLOCK <- 1/10
MAXDOMINANTBLOCK <- 1/4
INTRONWIDTH <- 1/10
EXONWIDTH <- 1/16
UTRWIDTH <- 1/12
PLOTMARGIN <- 16
##>

##################################################
#### remove chromosome from the input profile
noChromosome <- function(v) {
    return (as.numeric(v[2:length(v)]))
}
##################################################

##################################################
#### max coverage on a no-chromosome profile
maxCoverage <- function(v) {
    return (max(v[-c(1,2)]))
}
##################################################

##################################################
#### geometric mean
gmean <- function(x) {
    len <- length(x)
    if (len==0) {
        return (0)
    }
    return (prod(x^(1/len)))
}
##################################################

##################################################
#### max quantile average
qtmax <- function(v) {
    sv <- sort(v, decreasing=T)
    i <- length(v)*MINDOMINANTBLOCK
    while (i < length(v)*MAXDOMINANTBLOCK && sv[i] < sv[i+1] + TANDESCEND) {
        i <- i + 1
    }
    
    return (mean(sv[1:(i-1)]))
}
##################################################

##################################################
#### TMM
TMM <- function(countData) {
  require('edgeR')
  calcNormFactors(countData, method='TMM')
}
##################################################

##################################################
#### UQ
UQ <- function(countData) {
  require('edgeR')
  calcNormFactors(countData, method='upperquartile')
}
##################################################

##################################################
#### RLE
RLE <- function(countData) {
  require('edgeR')
  calcNormFactors(countData, method='RLE')
}
##################################################

##################################################
#### TC
TC <- function(countData) {
  cs <- colSums(countData)
  return (cs/gmean(cs))
}
##################################################

##################################################
#### Median
MED <- function(countData) {
  cs <- apply(countData, 2, median)
  return (cs/gmean(cs))
}
##################################################

##################################################
#### DESeq
DESeq <- function(countData) {
  require('DESeq')
  condition = factor(c(rep('A',NUMREP1), rep('B',NUMREP2)))
  cds = newCountDataSet(countData, condition)
  cds = estimateSizeFactors( cds )
  
  return (sizeFactors(cds))
}
##################################################

##################################################
#### calculate size factors for normalization
DominantBlock <- function(countData) {
  nr <- nrow(countData)
  nc <- ncol(countData)
  scd <- matrix(NA, nr, nc)
  print(dim(countData))
  ##< find size of high count data
  tmpCountData <- NULL
  maxlen <- LENTHRESHOLD
  
  for (j in 1:nc) {
    tmpCountData[[length(tmpCountData)+1]] <- countData[,j]
    maxCountData <- max(tmpCountData[[length(tmpCountData)]])
    tmpCountData[[length(tmpCountData)]] <- tmpCountData[[length(tmpCountData)]][tmpCountData[[length(tmpCountData)]] >= LOWCOVTHRESHOLD*maxCountData]
    maxlen <- max(maxlen, length(tmpCountData[[length(tmpCountData)]]))
  }
  ##>
  
  ##< sort count data
  for (j in 1:nc) {
    scd[,j] <- sort(countData[,j], decreasing=TRUE)
  }
  ##>
  
  ##< take only high sorted count data
  scd <- scd[1:maxlen,]
  ##>

  ##< find descending position
  i <- max(LENTHRESHOLD, maxlen*MINDOMINANTBLOCK)
  breakFlag <- FALSE
  while (i < maxlen*MAXDOMINANTBLOCK) {
    for (j in 1:nc) {
      if (scd[i,j] > scd[i+1,j] + TANDESCEND) {
        breakFlag <- TRUE
        break
      }
    }
    if (breakFlag) {
      break
    }
    i <- i + 1
  }
  ##>

  ##< calculate size factors
  scd <- scd[1:(i-1),]
  sf <- NULL
  sf <- apply(scd, 2, mean)
  
  ##< if a dominant block has less than EPSILONREAD reads
  if (length(sf[sf<=EPSILONREAD]) > 0) {
    return (NULL)
  }
  ##>
  #sf <- sf/mean(countMean[1:(i-1)])
  ##< normalize
  sf <- sf/gmean(sf)
  ##>
  
  return (sf)
}
##################################################

files1 = list('A1.cur', 'A2.cur', 'A3.cur')
files2 = list('B1.cur', 'B2.cur', 'B3.cur')

NUMGENE = 4
NUMREP1 = length(files1)
NUMREP2 = length(files2)
LABEL = c('A','B','C','D')
EXPECT1 = c("(3 3 3)", "(1 1 1)", "(1 1 1)", "(1 1 1)")
EXPECT2 = c("(1/3 1/3 1/3)", "(1 1 1)", "(1 1 1)", "(1 1 1)")
TITLE = c("EXPECTED = DB = DESeq = TC = MED; TMM, UQ, RLE fail",
          "EXPECTED = DB = TMM ~ UQ ~ TC; DESeq, RLE, MED fail",
          "EXPECTED = DB = DESeq; others fail",
          "EXPECTED = DB; others fail")

con1 = lapply(files1, function(x) {
  file(x, open="r")
})
con2 = lapply(files2, function(x) {
  file(x, open='r')
})
pdf(file=paste('rawCurve', '.pdf', sep=''), width=6, height=10)
par(mfrow=c(NUMGENE,1))
par(oma=c(0,1,0,0))
for (i in 1:NUMGENE) {
  vect1 = lapply(con1, function(x) {
    return ((strsplit(readLines(x, n=1), "\t"))[[1]])
  })
  vect2 = lapply(con2, function(x) {
    return ((strsplit(readLines(x, n=1), "\t"))[[1]])
  })
  
  vect1 <- lapply(vect1, noChromosome)
  vect2 <- lapply(vect2, noChromosome)
  
  lengene <- vect1[[1]][2] - vect1[[1]][1] + 1
  
  curve1 <- NULL
  curve2 <- NULL
  for (j in 1:NUMREP1) {
    curve1 <- rbind(curve1, vect1[[j]][-c(1,2)])
    if (lengene != length(vect1[[j]])-2) {
      print ("Unequal profile lengths")
      print(lengene)
      print(length(vect1[[j]]))
    }
  }
  for (j in 1:NUMREP2) {
    curve2 <- rbind(curve2, vect2[[j]][-c(1,2)])
    if (lengene != length(vect2[[j]])-2) {
      print ("Unequal profile lengths")
      print(lengene)
      print(length(vect1[[j]]))
    }
  }
  countData = t(rbind(curve1,curve2))
  #print(tail(countData))
  sft = rbind(DominantBlock(countData),
              DESeq(countData),
              TMM(countData),
              UQ(countData),
              RLE(countData),
              TC(countData),
              MED(countData))
  rownames(sft) = c('DB', 'DESeq', 'TMM', 'UQ', 'RLE', 'TC', 'MED')
  colors = c("red", "blue", "darkgreen", "deepskyblue", "gold", "magenta", "darkviolet")
  print(sft)
  
  colrep = c(2,2,2,4,4,4)
  ltyrep = c(1,4,5,1,4,5)
  plot(countData[,1], main=TITLE[i], xlim=c(-290,1340), xlab='Transcription Unit (positions)', ylab='RNA-seq read coverage', 
       type='l', lty=ltyrep[1], lwd=2, col=colrep[1], ylim=c(-90,max(countData)))
  sapply(2:6, function(k) {
    points(countData[,k], type='l', lty=ltyrep[k], lwd=1, col=colrep[k])
  })
  rect(0,-100,1000,-50)
  mtext(LABEL[i],side=3,at=-450,line=1,cex=1.27,las=1,font=2)
  text(1230, 550, 'EXPECTED:')
  text(1230, 400, EXPECT1[i], col=2)
  text(1230, 300, EXPECT2[i], col=4)
  legend('topright', c('experiment 1','experiment 2'), lty=1, col=c(2,4), cex=1)
  legend('bottomleft', c('replicate 1', 'replicate 2', 'replicate 3'), lty=ltyrep, cex=1)
  
#   pdf(file=paste('normalizedCurve', i, '.pdf', sep=''), width=18, height=6)
#   plot(countData[,1], type='l', lty=1, lwd=2, col=1, ylim=c(0,max(countData)))
#   sapply(4, function(k) {
#     points(countData[,k], type='l', lty=1, lwd=1, col=1)
#   })
#   sapply(1:(nrow(sft)), function(h) {
#     sapply(1, function(k) {
#       points(countData[,k]/sft[h,k], type='l', lty=1, lwd=2, col=colors[h])
#     })
#     sapply(4, function(k) {
#       points(countData[,k]/sft[h,k], type='l', lty=1, lwd=1, col=colors[h])
#     })
#   })
#   dev.off()
}
dev.off()
sapply(1:length(files1), function(i) {close(con1[[i]])})
sapply(1:length(files2), function(i) {close(con2[[i]])})

